package Id;
import javacard.framework.*;

public class Id extends Applet
{
  final static byte DEF_CLA = (byte) 0x80;

  final static byte DEF_SE_UNLOCK = (byte) 0x20;
// need pin unlock
  final static byte DEF_SE_GET_STATE = (byte) 0x21;
  final static byte DEF_SE_SET_STATE = (byte) 0x22;
  final static byte DEF_SE_GET_FREE_ID = (byte) 0x23;
  final static byte DEF_SE_SET_SELECTED_ID = (byte) 0x24;
  final static byte DEF_SE_GET_ID = (byte) 0x25;
  final static byte DEF_SE_SET_ID = (byte) 0x26;

// based on lock state
  final static byte DEF_RF_GET_UID = (byte) 0x60;
  final static byte DEF_RF_GET_PASS = (byte) 0x61;
  final static byte DEF_RF_GET_OTP = (byte) 0x62;
  final static byte DEF_RF_GET_NAME = (byte) 0x63;
  final static byte DEF_RF_GET_ADDR = (byte) 0x64;
  final static byte DEF_RF_GET_EMAIL = (byte) 0x65;

// lock states
  final static byte DEF_ONLEAVE_AUTOLOCK = (byte) 0x00;
  final static byte DEF_ONLEAVE_UNLOCK = (byte) 0x01;
  final static byte DEF_ONLEAVE_LOCK = (byte) 0x02;

// P2
  final static byte DEF_TAG_USED = (byte) 0x00;
  final static byte DEF_TAG_UID = (byte) 0x01;
  final static byte DEF_TAG_PASS = (byte) 0x02;
  final static byte DEF_TAG_OTP = (byte) 0x03;
  final static byte DEF_TAG_NAME = (byte) 0x04;
  final static byte DEF_TAG_ADDR = (byte) 0x05;
  final static byte DEF_TAG_EMAIL = (byte) 0x06;

  private static final byte MAX_TAG_UID = 15;
  private static final byte MAX_TAG_PASS = 15;
  private static final byte MAX_TAG_OTP = 6;
  private static final byte MAX_TAG_NAME = 30;
  private static final byte MAX_TAG_ADDR = 50;
  private static final byte MAX_TAG_EMAIL = 30;
  private static final byte MAX_IDS = 5;

  private static final byte NO_ID = (byte) 0x7f;

  final static byte PIN_TRY_LIMIT = (byte) 0x03;
  final static byte MAX_PIN_SIZE = (byte) 0x08;

  final static short SW_VERIFICATION_FAILED = 0x6300;
  final static short SW_PIN_VERIFICATION_REQUIRED = 0x6301;
  final static short SW_NO_ID = 0x6302;
  final static short SW_LOCKED = 0x6303;

  OwnerPIN pin;
  byte unlocked;
  byte unlocked_uid;
  byte unlocked_pass;
  byte unlocked_otp;
  byte unlocked_name;
  byte unlocked_addr;
  byte unlocked_email;
  byte id_selected;

    byte[] id_used;
    byte[] id_uid;
    byte[] id_pass;
    byte[] id_otp;
    byte[] id_name;
    byte[] id_addr;
    byte[] id_email;

  private Id (byte[]bArray, short bOffset, byte bLength)
  {
    unlocked = DEF_ONLEAVE_LOCK;
    unlocked_uid = DEF_ONLEAVE_LOCK;
    unlocked_pass = DEF_ONLEAVE_LOCK;
    unlocked_otp = DEF_ONLEAVE_LOCK;
    unlocked_name = DEF_ONLEAVE_LOCK;
    unlocked_addr = DEF_ONLEAVE_LOCK;
    unlocked_email = DEF_ONLEAVE_LOCK;
    id_selected = NO_ID;
    id_used = new byte[MAX_TAG_UID * MAX_IDS];
    id_uid = new byte[MAX_TAG_UID * MAX_IDS];
    id_pass = new byte[MAX_TAG_PASS * MAX_IDS];
    id_otp = new byte[MAX_TAG_OTP * MAX_IDS];
    id_name = new byte[MAX_TAG_NAME * MAX_IDS];
    id_addr = new byte[MAX_TAG_ADDR * MAX_IDS];
    id_email = new byte[MAX_TAG_EMAIL * MAX_IDS];

    for (byte i = 0; i < MAX_IDS; i++)
      id_used[i] = 0;

    pin = new OwnerPIN (PIN_TRY_LIMIT, MAX_PIN_SIZE);

    byte iLen = bArray[bOffset];	// aid length
      bOffset = (short) (bOffset + iLen + 1);
    byte cLen = bArray[bOffset];	// info length
      bOffset = (short) (bOffset + cLen + 1);
    byte aLen = bArray[bOffset];	// applet data length
      pin.update (bArray, (short) (bOffset + 1), aLen);

    register ();
  }

  public static void install (byte[]bArray, short bOffset, byte bLength)
  {
    new Id (bArray, bOffset, bLength);
  }

  public boolean select ()
  {
    if (pin.getTriesRemaining () == 0)
      return false;
    return true;
  }

  public void deselect ()
  {
    pin.reset ();
  }

  public void process (APDU apdu)
  {
    byte[]buffer = apdu.getBuffer ();

    buffer[ISO7816.OFFSET_CLA] =
      (byte) (buffer[ISO7816.OFFSET_CLA] & (byte) 0xfc);

    if ((buffer[ISO7816.OFFSET_CLA] == 0) &&
	(buffer[ISO7816.OFFSET_INS] == (byte) (0xa4)))
      return;

    if (buffer[ISO7816.OFFSET_CLA] != DEF_CLA)
      ISOException.throwIt (ISO7816.SW_CLA_NOT_SUPPORTED);

/*
TODO:
if any locked->all locked, i.e. autolock doesn't leave partial unlocked
*/

    switch (buffer[ISO7816.OFFSET_INS])
    {
      case DEF_SE_UNLOCK:
	se_unlock (apdu);
	return;
      case DEF_SE_GET_STATE:
	se_get_state (apdu);
	return;
      case DEF_SE_SET_STATE:
	se_set_state (apdu);
	return;
      case DEF_SE_GET_FREE_ID:
	se_get_free_id (apdu);
	return;
      case DEF_SE_SET_SELECTED_ID:
	se_set_selected_id (apdu);
	return;
      case DEF_SE_GET_ID:
	se_get_id (apdu);
	return;
      case DEF_SE_SET_ID:
	se_set_id (apdu);
	return;
      case DEF_RF_GET_UID:
	rf_get_uid (apdu);
	return;
      case DEF_RF_GET_PASS:
	rf_get_pass (apdu);
	return;
      case DEF_RF_GET_OTP:
	rf_get_otp (apdu);
	return;
      case DEF_RF_GET_NAME:
	rf_get_name (apdu);
	return;
      case DEF_RF_GET_ADDR:
	rf_get_addr (apdu);
	return;
      case DEF_RF_GET_EMAIL:
	rf_get_email (apdu);
	return;

      default:
	ISOException.throwIt (ISO7816.SW_INS_NOT_SUPPORTED);
    }
  }

  private void se_unlock (APDU apdu)
  {
    byte[]buffer = apdu.getBuffer ();
    byte byteRead = (byte) (apdu.setIncomingAndReceive ());

    if (pin.check (buffer, ISO7816.OFFSET_CDATA, byteRead) == false)
      ISOException.throwIt (SW_VERIFICATION_FAILED);
  }

  private void se_get_state (APDU apdu)
  {
    if (!pin.isValidated ())
      ISOException.throwIt (SW_PIN_VERIFICATION_REQUIRED);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = 1;

    buffer[0] = unlocked;
    apdu.setOutgoingLength (le);
    apdu.sendBytes ((short) 0, le);
  }

  private void se_set_state (APDU apdu)
  {
    if (!pin.isValidated ())
      ISOException.throwIt (SW_PIN_VERIFICATION_REQUIRED);

    byte[]buffer = apdu.getBuffer ();
    byte byteRead = (byte) (apdu.setIncomingAndReceive ());

// TODO no data

    unlocked = buffer[ISO7816.OFFSET_CDATA + 0];
    unlocked_uid = unlocked;
    unlocked_pass = unlocked;
    unlocked_otp = unlocked;
    unlocked_name = unlocked;
    unlocked_addr = unlocked;
    unlocked_email = unlocked;
  }

  private void se_get_free_id (APDU apdu)
  {
    if (!pin.isValidated ())
      ISOException.throwIt (SW_PIN_VERIFICATION_REQUIRED);
    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = 1;
    byte id = NO_ID;

    for (byte i = 0; i < MAX_IDS; i++)
      if (id_used[i] == 0)
	id = i;

    buffer[0] = id;
    apdu.setOutgoingLength (le);
    apdu.sendBytes ((short) 0, le);
  }

  private void se_set_selected_id (APDU apdu)
  {
    if (!pin.isValidated ())
      ISOException.throwIt (SW_PIN_VERIFICATION_REQUIRED);

    byte[]buffer = apdu.getBuffer ();
    byte byteRead = (byte) (apdu.setIncomingAndReceive ());

// TODO no data

    id_selected = buffer[ISO7816.OFFSET_CDATA + 0];
  }

  private void se_get_id (APDU apdu)
  {
    if (!pin.isValidated ())
      ISOException.throwIt (SW_PIN_VERIFICATION_REQUIRED);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    byte id = buffer[ISO7816.OFFSET_P1];
    byte tag = buffer[ISO7816.OFFSET_P2];

    switch (tag)
    {
      case DEF_TAG_USED:
	le = 1;
	for (short i = 0; i < le; i++)
	  buffer[i] = id_used[(short) ((id * le) + i)];
	break;

      case DEF_TAG_UID:
	le = MAX_TAG_UID;
	for (short i = 0; i < le; i++)
	  buffer[i] = id_uid[(short) ((id * le) + i)];
	break;

      case DEF_TAG_PASS:
	le = MAX_TAG_PASS;
	for (short i = 0; i < le; i++)
	  buffer[i] = id_pass[(short) ((id * le) + i)];
	break;

      case DEF_TAG_OTP:
	le = MAX_TAG_OTP;
	for (short i = 0; i < le; i++)
	  buffer[i] = id_otp[(short) ((id * le) + i)];
	update_otp (id);
	break;

      case DEF_TAG_NAME:
	le = MAX_TAG_NAME;
	for (short i = 0; i < le; i++)
	  buffer[i] = id_name[(short) ((id * le) + i)];
	break;

      case DEF_TAG_ADDR:
	le = MAX_TAG_ADDR;
	for (short i = 0; i < le; i++)
	  buffer[i] = id_addr[(short) ((id * le) + i)];
	break;

      case DEF_TAG_EMAIL:
	le = MAX_TAG_EMAIL;
	for (short i = 0; i < le; i++)
	  buffer[i] = id_email[(short) ((id * le) + i)];
	break;
    }

    apdu.setOutgoingLength ((byte) (le));
    apdu.sendBytes ((short) 0, (short) (le));
  }

  private void se_set_id (APDU apdu)
  {
    if (!pin.isValidated ())
      ISOException.throwIt (SW_PIN_VERIFICATION_REQUIRED);

    byte[]buffer = apdu.getBuffer ();
    byte id = buffer[ISO7816.OFFSET_P1];
    byte tag = buffer[ISO7816.OFFSET_P2];
    byte byteRead = (byte) (apdu.setIncomingAndReceive ());
    short le = 1;

// TODO no data

    switch (tag)
    {
      case DEF_TAG_USED:
	le = 1;
	for (short i = 0; i < le; i++)
	  if (i < byteRead)
	    id_used[(short) ((id * le) + i)] =
	      buffer[(short) (ISO7816.OFFSET_CDATA + i)];
	  else
	    id_used[(short) ((id * le) + i)] = 0;
	break;

      case DEF_TAG_UID:
	le = MAX_TAG_UID;
	for (short i = 0; i < le; i++)
	  if (i < byteRead)
	    id_uid[(short) ((id * le) + i)] =
	      buffer[(short) (ISO7816.OFFSET_CDATA + i)];
	  else
	    id_uid[(short) ((id * le) + i)] = 0;
	break;

      case DEF_TAG_PASS:
	le = MAX_TAG_PASS;
	for (short i = 0; i < le; i++)
	  if (i < byteRead)
	    id_pass[(short) ((id * le) + i)] =
	      buffer[(short) (ISO7816.OFFSET_CDATA + i)];
	  else
	    id_pass[(short) ((id * le) + i)] = 0;
	break;

      case DEF_TAG_OTP:
	le = MAX_TAG_OTP;
	for (short i = 0; i < le; i++)
	  if (i < byteRead)
	    id_otp[(short) ((id * le) + i)] =
	      buffer[(short) (ISO7816.OFFSET_CDATA + i)];
	  else
	    id_otp[(short) ((id * le) + i)] = '0';
	break;

      case DEF_TAG_NAME:
	le = MAX_TAG_NAME;
	for (short i = 0; i < le; i++)
	  if (i < byteRead)
	    id_name[(short) ((id * le) + i)] =
	      buffer[(short) (ISO7816.OFFSET_CDATA + i)];
	  else
	    id_name[(short) ((id * le) + i)] = 0;
	break;

      case DEF_TAG_ADDR:
	le = MAX_TAG_ADDR;
	for (short i = 0; i < le; i++)
	  if (i < byteRead)
	    id_addr[(short) ((id * le) + i)] =
	      buffer[(short) (ISO7816.OFFSET_CDATA + i)];
	  else
	    id_addr[(short) ((id * le) + i)] = 0;
	break;

      case DEF_TAG_EMAIL:
	le = MAX_TAG_EMAIL;
	for (short i = 0; i < le; i++)
	  if (i < byteRead)
	    id_email[(short) ((id * le) + i)] =
	      buffer[(short) (ISO7816.OFFSET_CDATA + i)];
	  else
	    id_email[(short) ((id * le) + i)] = 0;
	break;
    }
  }

  private void rf_get_uid (APDU apdu)
  {
    if (unlocked_uid == DEF_ONLEAVE_LOCK)
      ISOException.throwIt (SW_LOCKED);
    if (id_selected == NO_ID)
      ISOException.throwIt (SW_NO_ID);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = MAX_TAG_UID;

    for (short i = 0; i < le; i++)
      buffer[i] = id_uid[(short) ((id_selected * MAX_TAG_UID) + i)];

    if (unlocked_uid == DEF_ONLEAVE_AUTOLOCK)
      unlocked_uid = DEF_ONLEAVE_LOCK;

    apdu.setOutgoingLength ((byte) (le));
    apdu.sendBytes ((short) 0, (short) (le));
  }

  private void rf_get_pass (APDU apdu)
  {
    if (unlocked_pass == DEF_ONLEAVE_LOCK)
      ISOException.throwIt (SW_LOCKED);
    if (id_selected == NO_ID)
      ISOException.throwIt (SW_NO_ID);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = MAX_TAG_PASS;

    for (short i = 0; i < le; i++)
      buffer[i] = id_pass[(short) ((id_selected * MAX_TAG_PASS) + i)];

    if (unlocked_pass == DEF_ONLEAVE_AUTOLOCK)
      unlocked_pass = DEF_ONLEAVE_LOCK;

    apdu.setOutgoingLength ((byte) (le));
    apdu.sendBytes ((short) 0, (short) (le));
  }

  private void rf_get_otp (APDU apdu)
  {
    if (unlocked_otp == DEF_ONLEAVE_LOCK)
      ISOException.throwIt (SW_LOCKED);
    if (id_selected == NO_ID)
      ISOException.throwIt (SW_NO_ID);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = MAX_TAG_OTP;

    for (short i = 0; i < le; i++)
      buffer[i] = id_otp[(short) ((id_selected * MAX_TAG_OTP) + i)];

    update_otp (id_selected);

    if (unlocked_otp == DEF_ONLEAVE_AUTOLOCK)
      unlocked_otp = DEF_ONLEAVE_LOCK;

    apdu.setOutgoingLength ((byte) (le));
    apdu.sendBytes ((short) 0, (short) (le));
  }

  private void rf_get_name (APDU apdu)
  {
    if (unlocked_name == DEF_ONLEAVE_LOCK)
      ISOException.throwIt (SW_LOCKED);
    if (id_selected == NO_ID)
      ISOException.throwIt (SW_NO_ID);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = MAX_TAG_NAME;

    for (short i = 0; i < le; i++)
      buffer[i] = id_name[(short) ((id_selected * MAX_TAG_NAME) + i)];

    if (unlocked_name == DEF_ONLEAVE_AUTOLOCK)
      unlocked_name = DEF_ONLEAVE_LOCK;

    apdu.setOutgoingLength ((byte) (le));
    apdu.sendBytes ((short) 0, (short) (le));
  }

  private void rf_get_addr (APDU apdu)
  {
    if (unlocked_addr == DEF_ONLEAVE_LOCK)
      ISOException.throwIt (SW_LOCKED);
    if (id_selected == NO_ID)
      ISOException.throwIt (SW_NO_ID);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = MAX_TAG_ADDR;

    for (short i = 0; i < le; i++)
      buffer[i] = id_addr[(short) ((id_selected * MAX_TAG_ADDR) + i)];

    if (unlocked_addr == DEF_ONLEAVE_AUTOLOCK)
      unlocked_addr = DEF_ONLEAVE_LOCK;

    apdu.setOutgoingLength ((byte) (le));
    apdu.sendBytes ((short) 0, (short) (le));
  }

  private void rf_get_email (APDU apdu)
  {
    if (unlocked_email == DEF_ONLEAVE_LOCK)
      ISOException.throwIt (SW_LOCKED);
    if (id_selected == NO_ID)
      ISOException.throwIt (SW_NO_ID);

    byte[]buffer = apdu.getBuffer ();
    short le = apdu.setOutgoing ();

    le = MAX_TAG_EMAIL;

    for (short i = 0; i < le; i++)
      buffer[i] = id_email[(short) ((id_selected * MAX_TAG_EMAIL) + i)];

    if (unlocked_email == DEF_ONLEAVE_AUTOLOCK)
      unlocked_email = DEF_ONLEAVE_LOCK;

    apdu.setOutgoingLength ((byte) (le));
    apdu.sendBytes ((short) 0, (short) (le));
  }

  private void update_otp (byte id)
  {
    byte i = MAX_TAG_OTP - 1;
    byte over = 0;

    do
    {
      if (id_otp[(short) (id * MAX_TAG_OTP + i)] != '9')
      {
	id_otp[(short) (id * MAX_TAG_OTP + i)]++;
	over = 0;
      }
      else
      {
	over = 1;
	id_otp[(short) (id * MAX_TAG_OTP + i)] = '0';
      }
      i--;
    }
    while ((over != 0) && (i > 0));
  }
}
